@@ -108,37 +108,31 @@ class CoerciveIntFlag(IntFlag):
     @classmethod
     def coerce(cls, value):
         """
-                Attempt to coerce `value` into a member of this enumeration.
-        <<<<<<< HEAD
-        =======
+        Attempt to coerce `value` into a member of this enumeration.
 
-        >>>>>>> Handling page breaks in tables
-                If value is already a member of this enumeration it is returned unchanged.
-                Otherwise, if it is a string, attempt to convert it (case insensitively, by
-                upcasing) as an enumeration name. Otherwise, if it is an int, attempt to
-                convert it as an enumeration value.
-        <<<<<<< HEAD
-                Otherwise, an exception is raised.
-                Args:
-                    value (IntEnum, str, int): the value to be coerced.
-        =======
+        If value is already a member of this enumeration it is returned unchanged.
+        Otherwise, if it is a string, attempt to convert it (case insensitively, by
+        upcasing) as an enumeration name. Otherwise, if it is an int, attempt to
+        convert it as an enumeration value.
+        Otherwise, an exception is raised.
 
-                Otherwise, an exception is raised.
+        Args:
+            value (IntEnum, str, int): the value to be coerced.
 
-                Args:
-                    value (IntEnum, str, int): the value to be coerced.
-
-        >>>>>>> Handling page breaks in tables
-                Raises:
-                    ValueError: if `value` is an int but not a member of this enumeration.
-                    ValueError: if `value` is a string but not a member by name.
-                    TypeError: if `value`'s type is neither a member of the enumeration nor an
-                        int or a string.
+        Raises:
+            ValueError: if `value` is an int but not a member of this enumeration.
+            ValueError: if `value` is a string but not a member by name.
+            TypeError: if `value`'s type is neither a member of the enumeration nor an
+                int or a string.
         """
         if isinstance(value, cls):
             return value
 
         if isinstance(value, str):
+            try:
+                return cls[value.upper()]
+            except KeyError:
+                pass
             try:
                 flags = cls[value[0].upper()]
                 for char in value[1:]:
@@ -205,9 +199,30 @@ class Align(CoerciveEnum):
         return super(cls, cls).coerce(value)
 
 
+class VAlign(CoerciveEnum):
+    """Defines how to vertically render text in a cell.
+    Default value is MIDDLE"""
+
+    M = intern("MIDDLE")
+    "Center text vertically"
+
+    T = intern("TOP")
+    "Place text at the top of the cell, but obey the cells padding"
+
+    B = intern("BOTTOM")
+    "Place text at the bottom of the cell, but obey the cells padding"
+
+    @classmethod
+    def coerce(cls, value):
+        if value == "":
+            return cls.M
+        return super(cls, cls).coerce(value)
+
+
 class TextEmphasis(CoerciveIntFlag):
     """
     Indicates use of bold / italics / underline.
+
     This enum values can be combined with & and | operators:
         style = B | I
     """
@@ -241,6 +256,24 @@ class TextEmphasis(CoerciveIntFlag):
         return super(cls, cls).coerce(value)
 
 
+class MethodReturnValue(CoerciveIntFlag):
+    """
+    Defines the return value(s) of a FPDF content-rendering method.
+
+    This enum values can be combined with & and | operators:
+        PAGE_BREAK | LINES
+    """
+
+    PAGE_BREAK = 1
+    "The method will return a boolean indicating if a page break occured"
+
+    LINES = 2
+    "The method will return a multi-lines array of strings, after performing word-wrapping"
+
+    HEIGHT = 4
+    "The method will return how much vertical space was used"
+
+
 class TableBordersLayout(CoerciveEnum):
     "Defines how to render table borders"
 
@@ -281,6 +314,25 @@ class TableCellFillMode(CoerciveEnum):
     COLUMNS = intern("COLUMNS")
     "Fill only table cells in odd columns"
 
+    def should_fill_cell(self, i, j):
+        if self is self.NONE:
+            return False
+        if self is self.ALL:
+            return True
+        if self is self.ROWS:
+            return bool(i % 2)
+        if self is self.COLUMNS:
+            return bool(j % 2)
+        raise NotImplementedError
+
+
+class TableSpan(CoerciveEnum):
+    ROW = intern("ROW")
+    "Mark this cell as a continuation of the previous row"
+
+    COL = intern("COL")
+    "Mark this cell as a continuation of the previous column"
+
 
 class RenderStyle(CoerciveEnum):
     "Defines how to render shapes"
@@ -381,6 +433,18 @@ class YPos(CoerciveEnum):
     "bottom page margin (end of printable area)"
 
 
+class Angle(CoerciveIntEnum):
+    "Direction values used for mirror transformations specifying the angle of mirror line"
+    NORTH = 90
+    EAST = 0
+    SOUTH = 270
+    WEST = 180
+    NORTHEAST = 45
+    SOUTHEAST = 315
+    SOUTHWEST = 225
+    NORTHWEST = 135
+
+
 class PageLayout(CoerciveEnum):
     "Specify the page layout shall be used when the document is opened"
 
@@ -820,3 +884,4 @@ class EncryptionMethod(Enum):
     NO_ENCRYPTION = 0
     RC4 = 1
     AES_128 = 2
+    AES_256 = 3
