@@ -1,3 +1,12 @@
+"""
+This module contains the serialization logic that produces a PDF document from a FPDF instance.
+Most of the code in this module is used when FPDF.output() is called.
+
+The contents of this module are internal to fpdf2, and not part of the public API.
+They may change at any time without prior warning or any deprecation period,
+in non-backward-compatible ways.
+"""
+
 # pylint: disable=protected-access
 import logging
 from collections import defaultdict, OrderedDict
@@ -7,6 +16,7 @@ from io import BytesIO
 from .annotations import PDFAnnotation
 from .enums import SignatureFlag
 from .errors import FPDFException
+from .image_datastructures import RasterImageInfo
 from .outline import build_outline_objs
 from .sign import Signature, sign_content
 from .syntax import (
@@ -22,7 +32,6 @@ from .syntax import create_dictionary_string as pdf_dict
 from .syntax import create_list_string as pdf_list
 from .syntax import iobj_ref as pdf_ref
 
-from fontTools import ttLib
 from fontTools import subset as ftsubset
 
 try:
@@ -70,31 +79,6 @@ class PDFFont(PDFObject):
         self.c_i_d_to_g_i_d_map = None
 
 
-class PDFFontDescriptor(PDFObject):
-    def __init__(
-        self,
-        ascent,
-        descent,
-        cap_height,
-        flags,
-        font_b_box,
-        italic_angle,
-        stem_v,
-        missing_width,
-    ):
-        super().__init__()
-        self.type = Name("FontDescriptor")
-        self.ascent = ascent
-        self.descent = descent
-        self.cap_height = cap_height
-        self.flags = flags
-        self.font_b_box = font_b_box
-        self.italic_angle = italic_angle
-        self.stem_v = stem_v
-        self.missing_width = missing_width
-        self.font_name = None
-
-
 class CIDSystemInfo(PDFObject):
     def __init__(self):
         super().__init__()
@@ -115,12 +99,12 @@ class PDFInfo(PDFObject):
         creation_date: PDFDate,
     ):
         super().__init__()
-        self.title = PDFString(title) if title else None
-        self.subject = PDFString(subject) if subject else None
-        self.author = PDFString(author) if author else None
-        self.keywords = PDFString(keywords) if keywords else None
-        self.creator = PDFString(creator) if creator else None
-        self.producer = PDFString(producer) if producer else None
+        self.title = PDFString(title, encrypt=True) if title else None
+        self.subject = PDFString(subject, encrypt=True) if subject else None
+        self.author = PDFString(author, encrypt=True) if author else None
+        self.keywords = PDFString(keywords, encrypt=True) if keywords else None
+        self.creator = PDFString(creator, encrypt=True) if creator else None
+        self.producer = PDFString(producer, encrypt=True) if producer else None
         self.creation_date = creation_date
 
 
@@ -546,61 +530,39 @@ class OutputProducer:
 
     def _add_fonts(self):
         font_objs_per_index = {}
-        for font in sorted(self.fpdf.fonts.values(), key=lambda font: font["i"]):
+        for font in sorted(self.fpdf.fonts.values(), key=lambda font: font.i):
             # Standard font
-            if font["type"] == "core":
+            if font.type == "core":
                 encoding = (
                     "WinAnsiEncoding"
-                    if font["name"] not in ("Symbol", "ZapfDingbats")
+                    if font.name not in ("Symbol", "ZapfDingbats")
                     else None
                 )
                 core_font_obj = PDFFont(
-                    subtype="Type1", base_font=font["name"], encoding=encoding
+                    subtype="Type1", base_font=font.name, encoding=encoding
                 )
                 self._add_pdf_obj(core_font_obj, "fonts")
-                font_objs_per_index[font["i"]] = core_font_obj
-            elif font["type"] == "TTF":
-                fontname = f"MPDFAA+{font['name']}"
-
-                # unicode_char -> new_code_char map for chars embedded in the PDF
-                uni_to_new_code_char = font["subset"].dict()
-
-                # why we delete 0-element?
-                del uni_to_new_code_char[0]
-
-                # ---- FONTTOOLS SUBSETTER ----
-                # recalcTimestamp=False means that it doesn't modify the "modified" timestamp in head table
-                # if we leave recalcTimestamp=True the tests will break every time
-                fonttools_font = ttLib.TTFont(
-                    file=font["ttffile"], recalcTimestamp=False, fontNumber=0, lazy=True
-                )
+                font_objs_per_index[font.i] = core_font_obj
+            elif font.type == "TTF":
+                fontname = f"MPDFAA+{font.name}"
 
                 # 1. get all glyphs in PDF
-                cmap = fonttools_font["cmap"].getBestCmap()
-                glyph_names = [
-                    cmap[unicode] for unicode in uni_to_new_code_char if unicode in cmap
-                ]
-                print(glyph_names)
+                glyph_names = font.subset.get_all_glyph_names()
 
-                missing_glyphs = [
-                    chr(unicode)
-                    for unicode in uni_to_new_code_char
-                    if unicode not in cmap
-                ]
-                if len(missing_glyphs) > 0:
+                if len(font.missing_glyphs) > 0:
                     LOGGER.warning(
                         "Font %s is missing the following glyphs: %s",
                         fontname,
-                        ", ".join(missing_glyphs),
+                        ", ".join(chr(x) for x in font.missing_glyphs),
                     )
 
                 # 2. make a subset
                 # notdef_outline=True means that keeps the white box for the .notdef glyph
                 # recommended_glyphs=True means that adds the .notdef, .null, CR, and space glyphs
                 options = ftsubset.Options(notdef_outline=True, recommended_glyphs=True)
-                # dropping the tables previous dropped in the old ttfonts.py file #issue 418
+                # dropping some tables that currently not used:
                 options.drop_tables += [
-                    "FFTM",  # FontForge Timestamp table - cf. https://github.com/PyFPDF/fpdf2/issues/600
+                    "FFTM",  # FontForge Timestamp table - cf. https://github.com/py-pdf/fpdf2/issues/600
                     "GDEF",  # Glyph Definition table = various glyph properties used in OpenType layout processing
                     "GPOS",  # Glyph Positioning table = precise control over glyph placement
                     #          for sophisticated text layout and rendering in each script and language system
@@ -608,10 +570,11 @@ class OutputProducer:
                     "MATH",  # Mathematical typesetting table = specific information necessary for math formula layout
                     "hdmx",  # Horizontal Device Metrics table, stores integer advance widths scaled to particular pixel sizes
                     #          for OpenType™ fonts with TrueType outlines
+                    "meta",  # metadata table
                 ]
                 subsetter = ftsubset.Subsetter(options)
                 subsetter.populate(glyphs=glyph_names)
-                subsetter.subset(fonttools_font)
+                subsetter.subset(font.ttfont)
 
                 # 3. make codeToGlyph
                 # is a map Character_ID -> Glyph_ID
@@ -619,18 +582,15 @@ class OutputProducer:
                 # this basically takes the old code of the character
                 # take the glyph associated with it
                 # and then associate to the new code the glyph associated with the old code
-                code_to_glyph = {}
-                for code, new_code_mapped in uni_to_new_code_char.items():
-                    # notdef is associated if no glyph was associated to the old code
-                    # it's not necessary to do this, it seems to be done by default
-                    glyph_name = cmap.get(code, ".notdef")
-                    code_to_glyph[new_code_mapped] = fonttools_font.getGlyphID(
-                        glyph_name
-                    )
+
+                code_to_glyph = {
+                    char_id: font.ttfont.getGlyphID(glyph.glyph_name)
+                    for glyph, char_id in font.subset.items()
+                }
 
                 # 4. return the ttfile
                 output = BytesIO()
-                fonttools_font.save(output)
+                font.ttfont.save(output)
 
                 output.seek(0)
                 ttfontstream = output.read()
@@ -641,15 +601,15 @@ class OutputProducer:
                     subtype="Type0", base_font=fontname, encoding="Identity-H"
                 )
                 self._add_pdf_obj(composite_font_obj, "fonts")
-                font_objs_per_index[font["i"]] = composite_font_obj
+                font_objs_per_index[font.i] = composite_font_obj
 
                 # A CIDFont whose glyph descriptions are based on
                 # TrueType font technology
                 cid_font_obj = PDFFont(
                     subtype="CIDFontType2",
                     base_font=fontname,
-                    d_w=font["desc"].missing_width,
-                    w=_tt_font_widths(font, max(uni_to_new_code_char)),
+                    d_w=font.desc.missing_width,
+                    w=_tt_font_widths(font),
                 )
                 self._add_pdf_obj(cid_font_obj, "fonts")
                 composite_font_obj.descendant_fonts = PDFArray([cid_font_obj])
@@ -659,17 +619,21 @@ class OutputProducer:
                 # character that each used 16-bit code belongs to. It
                 # allows searching the file and copying text from it.
                 bfChar = []
-                uni_to_new_code_char = font["subset"].dict()
-                for code, code_mapped in uni_to_new_code_char.items():
-                    if code > 0xFFFF:
+
+                def format_code(unicode):
+                    if unicode > 0xFFFF:
                         # Calculate surrogate pair
-                        code_high = 0xD800 | (code - 0x10000) >> 10
-                        code_low = 0xDC00 | (code & 0x3FF)
-                        bfChar.append(
-                            f"<{code_mapped:04X}> <{code_high:04X}{code_low:04X}>\n"
-                        )
-                    else:
-                        bfChar.append(f"<{code_mapped:04X}> <{code:04X}>\n")
+                        code_high = 0xD800 | (unicode - 0x10000) >> 10
+                        code_low = 0xDC00 | (unicode & 0x3FF)
+                        return f"{code_high:04X}{code_low:04X}"
+                    return f"{unicode:04X}"
+
+                for glyph, code_mapped in font.subset.items():
+                    if len(glyph.unicode) == 0:
+                        continue
+                    bfChar.append(
+                        f'<{code_mapped:04X}> <{"".join(format_code(code) for code in glyph.unicode)}>\n'
+                    )
 
                 to_unicode_obj = PDFContentStream(
                     "/CIDInit /ProcSet findresource begin\n"
@@ -700,7 +664,7 @@ class OutputProducer:
                 self._add_pdf_obj(cid_system_info_obj, "fonts")
                 cid_font_obj.c_i_d_system_info = cid_system_info_obj
 
-                font_descriptor_obj = font["desc"]
+                font_descriptor_obj = font.desc
                 font_descriptor_obj.font_name = Name(fontname)
                 self._add_pdf_obj(font_descriptor_obj, "fonts")
                 cid_font_obj.font_descriptor = font_descriptor_obj
@@ -724,11 +688,15 @@ class OutputProducer:
                 self._add_pdf_obj(font_file_cs_obj, "fonts")
                 font_descriptor_obj.font_file2 = font_file_cs_obj
 
+                font.close()
+
         return font_objs_per_index
 
     def _add_images(self):
         img_objs_per_index = {}
-        for img in sorted(self.fpdf.images.values(), key=lambda img: img["i"]):
+        for img in sorted(
+            self.fpdf.image_cache.images.values(), key=lambda img: img["i"]
+        ):
             if img["usages"] > 0:
                 img_objs_per_index[img["i"]] = self._add_image(img)
         return img_objs_per_index
@@ -742,7 +710,7 @@ class OutputProducer:
         if iccp_i in self.iccp_i_to_pdf_i:
             return self.iccp_i_to_pdf_i[iccp_i]
         iccp_content = None
-        for iccp_c, i in self.fpdf.icc_profiles.items():
+        for iccp_c, i in self.fpdf.image_cache.icc_profiles.items():
             if iccp_i == i:
                 iccp_content = iccp_c
                 break
@@ -767,10 +735,8 @@ class OutputProducer:
             iccp_pdf_i = self._ensure_iccp(info)
             color_space = PDFArray(["/ICCBased", str(iccp_pdf_i), str("0"), "R"])
         elif color_space == "DeviceCMYK":
-            decode = "[1 0 1 0 1 0 1 0]"
-            raise NotImplementedError(
-                "fpdf2 does not support DeviceCMYK ColorSpace yet - cf. issue #711"
-            )
+            if info["inverted"] is True:
+                decode = "[1 0 1 0 1 0 1 0]"
 
         decode_parms = f"<<{info['dp']} /BitsPerComponent {info['bpc']}>>"
         img_obj = PDFXObject(
@@ -958,7 +924,13 @@ class OutputProducer:
                 *ZOOM_CONFIGS[fpdf.zoom_mode],
             ]
         else:  # zoom_mode is a number, not one of the allowed strings:
-            zoom_config = ["/XYZ", "null", "null", str(fpdf.zoom_mode / 100)]
+            zoom_config = [
+                pdf_ref(first_page_obj.id),
+                "/XYZ",
+                "null",
+                "null",
+                str(fpdf.zoom_mode / 100),
+            ]
         catalog_obj.open_action = pdf_list(zoom_config)
         if struct_tree_root_obj:
             catalog_obj.mark_info = pdf_dict({"/Marked": "true"})
@@ -984,49 +956,69 @@ class OutputProducer:
             LOGGER.debug("- %s: %s", label, _sizeof_fmt(section_size))
 
 
-def _tt_font_widths(font, maxUni):
+def stream_content_for_raster_image(
+    info: RasterImageInfo,
+    x,
+    y,
+    w,
+    h,
+    keep_aspect_ratio=False,
+    scale=1,
+    pdf_height_to_flip=None,
+):
+    if keep_aspect_ratio:
+        x, y, w, h = info.scale_inside_box(x, y, w, h)
+    if pdf_height_to_flip:
+        stream_h = h
+        stream_y = pdf_height_to_flip - h - y
+    else:
+        stream_h = -h
+        stream_y = y + h
+    return (
+        f"q {w * scale:.2f} 0 0 {stream_h * scale:.2f}"
+        f" {x * scale:.2f} {stream_y * scale:.2f} cm"
+        f" /I{info['i']} Do Q"
+    )
+
+
+def _tt_font_widths(font):
     rangeid = 0
     range_ = {}
     range_interval = {}
     prevcid = -2
     prevwidth = -1
     interval = False
-    startcid = 1
-    cwlen = maxUni + 1
 
-    # for each character
-    subset = font["subset"].dict()
-    for cid in range(startcid, cwlen):
-        char_width = font["cw"][cid]
-        if "dw" not in font or (font["dw"] and char_width != font["dw"]):
-            cid_mapped = subset.get(cid)
-            if cid_mapped is None:
-                continue
-            if cid_mapped == (prevcid + 1):
-                if char_width == prevwidth:
-                    if char_width == range_[rangeid][0]:
-                        range_.setdefault(rangeid, []).append(char_width)
-                    else:
-                        range_[rangeid].pop()
-                        # new range
-                        rangeid = prevcid
-                        range_[rangeid] = [prevwidth, char_width]
-                    interval = True
-                    range_interval[rangeid] = True
+    # Glyphs sorted by mapped character id
+    glyphs = dict(sorted(font.subset.items(), key=lambda item: item[1]))
+
+    for glyph in glyphs:
+        cid_mapped = glyphs[glyph]
+        if cid_mapped == (prevcid + 1):
+            if glyph.glyph_width == prevwidth:
+                if glyph.glyph_width == range_[rangeid][0]:
+                    range_.setdefault(rangeid, []).append(glyph.glyph_width)
                 else:
-                    if interval:
-                        # new range
-                        rangeid = cid_mapped
-                        range_[rangeid] = [char_width]
-                    else:
-                        range_[rangeid].append(char_width)
-                    interval = False
+                    range_[rangeid].pop()
+                    # new range
+                    rangeid = prevcid
+                    range_[rangeid] = [prevwidth, glyph.glyph_width]
+                interval = True
+                range_interval[rangeid] = True
             else:
-                rangeid = cid_mapped
-                range_[rangeid] = [char_width]
+                if interval:
+                    # new range
+                    rangeid = cid_mapped
+                    range_[rangeid] = [glyph.glyph_width]
+                else:
+                    range_[rangeid].append(glyph.glyph_width)
                 interval = False
-            prevcid = cid_mapped
-            prevwidth = char_width
+        else:
+            rangeid = cid_mapped
+            range_[rangeid] = [glyph.glyph_width]
+            interval = False
+        prevcid = cid_mapped
+        prevwidth = glyph.glyph_width
     prevk = -1
     nextk = -1
     prevint = False
