@@ -1,4 +1,14 @@
-import math, re, warnings
+"""
+Utilities to parse SVG graphics into fpdf.drawing objects.
+
+The contents of this module are internal to fpdf2, and not part of the public API.
+They may change at any time without prior warning or any deprecation period,
+in non-backward-compatible ways.
+"""
+
+import logging, math, re, warnings
+from numbers import Number
+from typing import NamedTuple
 
 from fontTools.svgLib.path import parse_path
 from fontTools.pens.basePen import BasePen
@@ -19,8 +29,13 @@ from .drawing import (
     GraphicsContext,
     GraphicsStyle,
     PaintedPath,
+    ClippingPath,
     Transform,
 )
+from .image_datastructures import ImageCache, VectorImageInfo
+from .output import stream_content_for_raster_image
+
+LOGGER = logging.getLogger(__name__)
 
 __pdoc__ = {"force_nodocument": False}
 
@@ -155,6 +170,15 @@ def xmlns_lookup(space, *names):
     return result
 
 
+@force_nodocument
+def without_ns(qualified_tag):
+    """Remove the xmlns namespace from a qualified XML tag name"""
+    i = qualified_tag.index("}")
+    if i >= 0:
+        return qualified_tag[i + 1 :]
+    return qualified_tag
+
+
 shape_tags = xmlns_lookup(
     "svg", "rect", "circle", "ellipse", "line", "polyline", "polygon"
 )
@@ -303,8 +327,7 @@ def apply_styles(stylable, svg_element):
     for attr_name, converter in svg_attr_map.items():
         value = svg_element.attrib.get(attr_name)
         if value:
-            attr_name, value = converter(value)
-            setattr(stylable.style, attr_name, value)
+            setattr(stylable.style, *converter(value))
 
     # handle this separately for now
     opacity = svg_element.attrib.get("opacity")
@@ -323,21 +346,30 @@ class ShapeBuilder:
     """A namespace within which methods for converting basic shapes can be looked up."""
 
     @staticmethod
-    def new_path(tag):
+    def new_path(tag, clipping_path: bool = False):
         """Create a new path with the appropriate styles."""
         path = PaintedPath()
+        if clipping_path:
+            path = ClippingPath()
         apply_styles(path, tag)
-
         return path
 
     @classmethod
-    def rect(cls, tag):
+    def rect(cls, tag, clipping_path: bool = False):
         """Convert an SVG <rect> into a PDF path."""
         # svg rect is wound clockwise
-        x = float(tag.attrib.get("x", 0))
-        y = float(tag.attrib.get("y", 0))
-        width = resolve_length(tag.attrib.get("width", 0))
-        height = resolve_length(tag.attrib.get("height", 0))
+        x = resolve_length(tag.attrib.get("x", "0"))
+        y = resolve_length(tag.attrib.get("y", "0"))
+        width = tag.attrib.get("width", "0")
+        if width.endswith("%"):
+            width = Percent(width[:-1])
+        else:
+            width = resolve_length(width)
+        height = tag.attrib.get("height", "0")
+        if height.endswith("%"):
+            height = Percent(height[:-1])
+        else:
+            height = resolve_length(height)
         rx = tag.attrib.get("rx", "auto")
         ry = tag.attrib.get("ry", "auto")
 
@@ -367,25 +399,23 @@ class ShapeBuilder:
         if ry > (height / 2):
             ry = height / 2
 
-        path = cls.new_path(tag)
-
+        path = cls.new_path(tag, clipping_path)
         path.rectangle(x, y, width, height, rx, ry)
         return path
 
     @classmethod
-    def circle(cls, tag):
+    def circle(cls, tag, clipping_path: bool = False):
         """Convert an SVG <circle> into a PDF path."""
         cx = float(tag.attrib.get("cx", 0))
         cy = float(tag.attrib.get("cy", 0))
         r = float(tag.attrib["r"])
 
-        path = cls.new_path(tag)
-
+        path = cls.new_path(tag, clipping_path)
         path.circle(cx, cy, r)
         return path
 
     @classmethod
-    def ellipse(cls, tag):
+    def ellipse(cls, tag, clipping_path: bool = False):
         """Convert an SVG <ellipse> into a PDF path."""
         cx = float(tag.attrib.get("cx", 0))
         cy = float(tag.attrib.get("cy", 0))
@@ -393,7 +423,7 @@ class ShapeBuilder:
         rx = tag.attrib.get("rx", "auto")
         ry = tag.attrib.get("ry", "auto")
 
-        path = cls.new_path(tag)
+        path = cls.new_path(tag, clipping_path)
 
         if (rx == ry == "auto") or (rx == 0) or (ry == 0):
             return path
@@ -418,34 +448,24 @@ class ShapeBuilder:
         y2 = float(tag.attrib["y2"])
 
         path = cls.new_path(tag)
-
         path.move_to(x1, y1)
         path.line_to(x2, y2)
-
         return path
 
     @classmethod
     def polyline(cls, tag):
         """Convert an SVG <polyline> into a PDF path."""
-        points = tag.attrib["points"]
-
         path = cls.new_path(tag)
-
-        points = "M" + points
+        points = "M" + tag.attrib["points"]
         svg_path_converter(path, points)
-
         return path
 
     @classmethod
-    def polygon(cls, tag):
+    def polygon(cls, tag, clipping_path: bool = False):
         """Convert an SVG <polygon> into a PDF path."""
-        points = tag.attrib["points"]
-
-        path = cls.new_path(tag)
-
-        points = "M" + points + "Z"
+        path = cls.new_path(tag, clipping_path)
+        points = "M" + tag.attrib["points"] + "Z"
         svg_path_converter(path, points)
-
         return path
 
 
@@ -459,7 +479,7 @@ def convert_transforms(tfstr):
     # harder to support.
     # https://drafts.csswg.org/css-transforms/#two-d-transform-functions
     parsed = TRANSFORM_GETTER.findall(tfstr)
-
+    # pylint: disable=redefined-loop-name
     transform = Transform.identity()
     for tf_type, args in parsed:
         if tf_type == "matrix":
@@ -633,7 +653,8 @@ class SVGObject:
         with open(filename, "r", encoding=encoding) as svgfile:
             return cls(svgfile.read(), *args, **kwargs)
 
-    def __init__(self, svg_text):
+    def __init__(self, svg_text, image_cache: ImageCache = None):
+        self.image_cache = image_cache  # Needed to render images
         self.cross_references = {}
 
         # disabling bandit rule as we use defusedxml:
@@ -645,6 +666,12 @@ class SVGObject:
         self.extract_shape_info(svg_tree)
         self.convert_graphics(svg_tree)
 
+    @force_nodocument
+    def update_xref(self, key, referenced):
+        if key:
+            key = "#" + key if not key.startswith("#") else key
+            self.cross_references[key] = referenced
+
     @force_nodocument
     def extract_shape_info(self, root_tag):
         """Collect shape info from the given SVG."""
@@ -816,6 +843,7 @@ class SVGObject:
             debug_stream (io.TextIO): the stream to which rendering debug info will be
                 written.
         """
+        self.image_cache = pdf.image_cache  # Needed to render images
         _, _, path = self.transform_to_page_viewport(pdf)
 
         old_x, old_y = pdf.x, pdf.y
@@ -837,9 +865,24 @@ class SVGObject:
         for child in defs:
             if child.tag in xmlns_lookup("svg", "g"):
                 self.build_group(child)
-            if child.tag in xmlns_lookup("svg", "path"):
+            elif child.tag in xmlns_lookup("svg", "path"):
                 self.build_path(child)
-            # We could/should also support <defs> that are rect, circle, ellipse, line, polyline, polygon...
+            elif child.tag in xmlns_lookup("svg", "image"):
+                self.build_image(child)
+            elif child.tag in shape_tags:
+                self.build_shape(child)
+            elif child.tag in xmlns_lookup("svg", "clipPath"):
+                try:
+                    clip_id = child.attrib["id"]
+                except KeyError:
+                    clip_id = None
+                for child_ in child:
+                    self.build_clipping_path(child_, clip_id)
+            else:
+                LOGGER.warning(
+                    "Ignoring unsupported SVG tag: <%s> (contributions are welcome to add support for it)",
+                    without_ns(child.tag),
+                )
 
     # this assumes xrefs only reference already-defined ids.
     # I don't know if this is required by the SVG spec.
@@ -849,7 +892,7 @@ class SVGObject:
         pdf_group = GraphicsContext()
         apply_styles(pdf_group, xref)
 
-        for candidate in xmlns_lookup("xlink", "href"):
+        for candidate in xmlns_lookup("xlink", "href", "id"):
             try:
                 ref = xref.attrib[candidate]
                 break
@@ -881,22 +924,32 @@ class SVGObject:
             pdf_group = GraphicsContext()
             apply_styles(pdf_group, group)
 
+        # handle defs before anything else
+        for child in [
+            child for child in group if child.tag in xmlns_lookup("svg", "defs")
+        ]:
+            self.handle_defs(child)
+
         for child in group:
             if child.tag in xmlns_lookup("svg", "defs"):
                 self.handle_defs(child)
-            if child.tag in xmlns_lookup("svg", "g"):
+            elif child.tag in xmlns_lookup("svg", "g"):
                 pdf_group.add_item(self.build_group(child))
-            if child.tag in xmlns_lookup("svg", "path"):
+            elif child.tag in xmlns_lookup("svg", "path"):
                 pdf_group.add_item(self.build_path(child))
             elif child.tag in shape_tags:
-                pdf_group.add_item(getattr(ShapeBuilder, shape_tags[child.tag])(child))
-            if child.tag in xmlns_lookup("svg", "use"):
+                pdf_group.add_item(self.build_shape(child))
+            elif child.tag in xmlns_lookup("svg", "use"):
                 pdf_group.add_item(self.build_xref(child))
+            elif child.tag in xmlns_lookup("svg", "image"):
+                pdf_group.add_item(self.build_image(child))
+            else:
+                LOGGER.warning(
+                    "Ignoring unsupported SVG tag: <%s> (contributions are welcome to add support for it)",
+                    without_ns(child.tag),
+                )
 
-        try:
-            self.cross_references["#" + group.attrib["id"]] = pdf_group
-        except KeyError:
-            pass
+        self.update_xref(group.attrib.get("id"), pdf_group)
 
         return pdf_group
 
@@ -905,15 +958,138 @@ class SVGObject:
         """Convert an SVG <path> tag into a PDF path object."""
         pdf_path = PaintedPath()
         apply_styles(pdf_path, path)
-
-        svg_path = path.attrib.get("d", None)
-
+        self.apply_clipping_path(pdf_path, path)
+        svg_path = path.attrib.get("d")
         if svg_path is not None:
             svg_path_converter(pdf_path, svg_path)
-
-        try:
-            self.cross_references["#" + path.attrib["id"]] = pdf_path
-        except KeyError:
-            pass
-
+        self.update_xref(path.attrib.get("id"), pdf_path)
         return pdf_path
+
+    @force_nodocument
+    def build_shape(self, shape):
+        """Convert an SVG shape tag into a PDF path object. Necessary to make xref (because ShapeBuilder doesn't have access to this object.)"""
+        shape_builder = getattr(ShapeBuilder, shape_tags[shape.tag])
+        shape_path = shape_builder(shape)
+        self.apply_clipping_path(shape_path, shape)
+        self.update_xref(shape.attrib.get("id"), shape_path)
+        return shape_path
+
+    @force_nodocument
+    def build_clipping_path(self, shape, clip_id):
+        if shape.tag in shape_tags:
+            shape_builder = getattr(ShapeBuilder, shape_tags[shape.tag])
+            clipping_path_shape = shape_builder(shape, True)
+        elif shape.tag in xmlns_lookup("svg", "path"):
+            clipping_path_shape = PaintedPath()
+            apply_styles(clipping_path_shape, shape)
+            svg_path = shape.attrib.get("d")
+            if svg_path is not None:
+                svg_path_converter(clipping_path_shape, svg_path)
+        else:
+            LOGGER.warning(
+                "Ignoring unsupported <clipPath> child tag: <%s> (contributions are welcome to add support for it)",
+                without_ns(shape.tag),
+            )
+            return
+        self.update_xref(clip_id, clipping_path_shape)
+
+    @force_nodocument
+    def apply_clipping_path(self, stylable, svg_element):
+        clipping_path = svg_element.attrib.get("clip-path")
+        if clipping_path:
+            clipping_path_id = re.search(r"url\((\#\w+)\)", clipping_path)
+            stylable.clipping_path = self.cross_references[clipping_path_id[1]]
+
+    @force_nodocument
+    def build_image(self, image):
+        href = None
+        for key in xmlns_lookup("xlink", "href"):
+            if key in image.attrib:
+                href = image.attrib[key]
+                break
+        if not href:
+            raise ValueError("<image> is missing a href attribute")
+        width = float(image.attrib.get("width", 0))
+        height = float(image.attrib.get("height", 0))
+        if "preserveAspectRatio" in image.attrib:
+            LOGGER.warning(
+                '"preserveAspectRatio" defined on <image> is currently not supported (contributions are welcome to add support for it)'
+            )
+        if "style" in image.attrib:
+            LOGGER.warning(
+                '"style" defined on <image> is currently not supported (contributions are welcome to add support for it)'
+            )
+        if "transform" in image.attrib:
+            LOGGER.warning(
+                '"transform" defined on <image> is currently not supported (contributions are welcome to add support for it)'
+            )
+        # Note: at this moment, self.image_cache is not set yet:
+        svg_image = SVGImage(
+            href=href,
+            x=float(image.attrib.get("x", "0")),
+            y=float(image.attrib.get("y", "0")),
+            width=width,
+            height=height,
+            svg_obj=self,
+        )
+        self.update_xref(image.attrib.get("id"), svg_image)
+        return svg_image
+
+
+class SVGImage(NamedTuple):
+    href: str
+    x: Number
+    y: Number
+    width: Number
+    height: Number
+    svg_obj: SVGObject
+
+    def __deepcopy__(self, _memo):
+        # Defining this method is required to avoid the .svg_obj reference to be cloned:
+        return SVGImage(
+            href=self.href,
+            x=self.x,
+            y=self.y,
+            width=self.width,
+            height=self.height,
+            svg_obj=self.svg_obj,
+        )
+
+    @force_nodocument
+    def render(self, _gsd_registry, _style, last_item, initial_point):
+        image_cache = self.svg_obj and self.svg_obj.image_cache
+        if not image_cache:
+            raise AssertionError(
+                "fpdf2 bug - Cannot render a raster image without a SVGObject.image_cache"
+            )
+
+        # We lazy-import this function to circumvent a circular import problem:
+        # pylint: disable=cyclic-import,import-outside-toplevel
+        from .image_parsing import preload_image
+
+        _, _, info = preload_image(image_cache, self.href)
+        if isinstance(info, VectorImageInfo):
+            LOGGER.warning(
+                "Inserting .svg vector graphics in <image> tags is currently not supported (contributions are welcome to add support for it)"
+            )
+            return "", last_item, initial_point
+        w, h = info.size_in_document_units(self.width, self.height)
+        stream_content = stream_content_for_raster_image(
+            info=info,
+            x=self.x,
+            y=self.y,
+            w=w,
+            h=h,
+            keep_aspect_ratio=True,
+        )
+        return stream_content, last_item, initial_point
+
+    @force_nodocument
+    def render_debug(
+        self, gsd_registry, style, last_item, initial_point, debug_stream, _pfx
+    ):
+        stream_content, last_item, initial_point = self.render(
+            gsd_registry, style, last_item, initial_point
+        )
+        debug_stream.write(f"{self.href} rendered as: {stream_content}\n")
+        return stream_content, last_item, initial_point
